@@ -17,6 +17,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <unistd.h>
 #include <string.h>
 #include <signal.h>
 #include <syslog.h>
@@ -33,9 +34,15 @@
 #include "common.h"
 #include "log.h"
 #include "netconf.h"
-#include "netconf_server.h"
-#include "netconf_server_ssh.h"
-#include "netconf_server_tls.h"
+#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
+# include "netconf_server.h"
+#endif
+#ifdef NC_ENABLED_SSH
+# include "netconf_server_ssh.h"
+#endif
+#ifdef NC_ENABLED_TLS
+# include "netconf_server_tls.h"
+#endif
 #include "netconf_acm.h"
 #include "netconf_monitoring.h"
 #include "netconf_nmda.h"
@@ -211,6 +218,8 @@ cleanup:
     return ret;
 }
 
+#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
+
 static int
 np2srv_dummy_cb(sr_session_ctx_t *UNUSED(session), const char *UNUSED(module_name), const char *UNUSED(xpath),
         sr_event_t UNUSED(event), uint32_t UNUSED(request_id), void *UNUSED(private_data))
@@ -218,6 +227,8 @@ np2srv_dummy_cb(sr_session_ctx_t *UNUSED(session), const char *UNUSED(module_nam
     return SR_ERR_OK;
 }
 
+#endif
+
 static void
 np2srv_del_session_cb(struct nc_session *session)
 {
@@ -234,21 +245,7 @@ np2srv_del_session_cb(struct nc_session *session)
     /* stop sysrepo session (also stop any sysrepo notification subscriptions) */
     sr_sess = nc_session_get_data(session);
     sr_session_stop(sr_sess);
-
-    switch (nc_session_get_ti(session)) {
-#ifdef NC_ENABLED_SSH
-    case NC_TI_LIBSSH:
-#endif
-#ifdef NC_ENABLED_TLS
-    case NC_TI_OPENSSL:
-#endif
-#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
-        ncm_session_del(session);
-        break;
-#endif
-    default:
-        break;
-    }
+    ncm_session_del(session);
 
     if ((mod = ly_ctx_get_module(sr_get_context(np2srv.sr_conn), "ietf-netconf-notifications", NULL, 1))) {
         /* generate ietf-netconf-notification's netconf-session-end event for sysrepo */
@@ -344,9 +341,28 @@ err_access_denied:
         /* fallthrough */
     default:
         if (strstr(message, "authorization failed")) {
+            // access-denied
             goto err_access_denied;
-        } else if (strstr(message, "is already locked")) {
+        } else if (strstr(message, "is already locked") ||
+                   strstr(message, "Module \"yang\" is locked by session")) {
+            // lock-denied
             goto err_lock_denied;
+        } else if (strstr(message, "is locked by session")) {
+            // in-use
+            e = nc_err(NC_ERR_IN_USE, NC_ERR_TYPE_PROT);
+            nc_err_set_msg(e, message, "en");
+            if (xpath) {
+                nc_err_set_path(e, xpath);
+            }
+            break;
+        } else if (strstr(message, "already exists")) {
+            // data-exists
+            e = nc_err(NC_ERR_DATA_EXISTS);
+            nc_err_set_msg(e, message, "en");
+            if (xpath) {
+                nc_err_set_path(e, xpath);
+            }
+            break;
         } else if (strstr(message, "Source and target")) {
             e = nc_err(NC_ERR_INVALID_VALUE, NC_ERR_TYPE_PROT);
             nc_err_set_msg(e, message, "en");
@@ -424,8 +440,8 @@ np2srv_rpc_cb(struct lyd_node *rpc, struct nc_session *ncs)
     /* get this user session with its NC id (but not user name) */
     sr_sess = nc_session_get_data(ncs);
 
-    /* sysrepo API */
-    rc = sr_rpc_send_tree(sr_sess, rpc, NP2SRV_RPC_TIMEOUT, &output);
+    /* sysrepo API, use the default timeout or slightly higher than the configured one */
+    rc = sr_rpc_send_tree(sr_sess, rpc, np2srv.sr_timeout ? np2srv.sr_timeout + 2000 : 0, &output);
     if (rc != SR_ERR_OK) {
         ERR("Failed to send an RPC (%s).", sr_strerror(rc));
         goto cleanup;
@@ -441,7 +457,13 @@ np2srv_rpc_cb(struct lyd_node *rpc, struct nc_session *ncs)
     }
     if (child) {
         /* get with-defaults mode */
-        nodeset = lyd_find_path(rpc, "ietf-netconf-with-defaults:with-defaults");
+        if (!strcmp(lyd_node_module(rpc)->name, "ietf-netconf")) {
+            /* augment */
+            nodeset = lyd_find_path(rpc, "ietf-netconf-with-defaults:with-defaults");
+        } else {
+            /* grouping */
+            nodeset = lyd_find_path(rpc, "with-defaults");
+        }
         if (nodeset->number) {
             leaf = (struct lyd_node_leaf_list *)nodeset->set.d[0];
             if (!strcmp(leaf->value_str, "report-all")) {
@@ -682,13 +704,17 @@ server_init(void)
     /* set libnetconf2 global PRC callback */
     nc_set_global_rpc_clb(np2srv_rpc_cb);
 
+#ifdef NC_ENABLED_SSH
     /* set libnetconf2 SSH callbacks */
     nc_server_ssh_set_hostkey_clb(np2srv_hostkey_cb, NULL, NULL);
     nc_server_ssh_set_pubkey_auth_clb(np2srv_pubkey_auth_cb, NULL, NULL);
+#endif
 
+#ifdef NC_ENABLED_TLS
     /* set libnetconf2 TLS callbacks */
     nc_server_tls_set_server_cert_clb(np2srv_cert_cb, NULL, NULL);
     nc_server_tls_set_trusted_cert_list_clb(np2srv_cert_list_cb, NULL, NULL);
+#endif
 
     /* UNIX socket */
     if (np2srv.unix_path) {
@@ -799,9 +825,13 @@ server_data_subscribe(void)
      * ietf-netconf-server
      */
     mod_name = "ietf-netconf-server";
+
+#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
     xpath = "/ietf-netconf-server:netconf-server/listen/idle-timeout";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_idle_timeout_cb);
+#endif
 
+#ifdef NC_ENABLED_SSH
     /* subscribe for server SSH listen configuration changes */
     xpath = "/ietf-netconf-server:netconf-server/listen/endpoint/ssh";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_endpt_ssh_cb);
@@ -820,7 +850,9 @@ server_data_subscribe(void)
     /* subscribe for providing SSH operational data */
     xpath = "/ietf-netconf-server:netconf-server/listen/endpoint/ssh/ssh-server-parameters/client-authentication/users";
     SR_OPER_SUBSCR(mod_name, xpath, np2srv_endpt_ssh_auth_users_oper_cb);
+#endif
 
+#ifdef NC_ENABLED_TLS
     /* subscribe for server TLS listen configuration changes */
     xpath = "/ietf-netconf-server:netconf-server/listen/endpoint/tls";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_endpt_tls_cb);
@@ -836,7 +868,9 @@ server_data_subscribe(void)
 
     xpath = "/ietf-netconf-server:netconf-server/listen/endpoint/tls/tls-server-parameters/client-authentication/cert-maps";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_endpt_tls_client_ctn_cb);
+#endif
 
+#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
     /* subscribe for generic Call Home configuration changes */
     xpath = "/ietf-netconf-server:netconf-server/call-home/netconf-client";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_ch_client_cb);
@@ -846,7 +880,9 @@ server_data_subscribe(void)
 
     xpath = "/ietf-netconf-server:netconf-server/call-home/netconf-client/reconnect-strategy";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_ch_reconnect_strategy_cb);
+#endif
 
+#ifdef NC_ENABLED_SSH
     /* subscribe for server SSH Call Home configuration changes */
     xpath = "/ietf-netconf-server:netconf-server/call-home/netconf-client/endpoints/endpoint/ssh";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_ch_client_endpt_ssh_cb);
@@ -861,7 +897,9 @@ server_data_subscribe(void)
     xpath = "/ietf-netconf-server:netconf-server/call-home/netconf-client/endpoints/endpoint/ssh/ssh-server-parameters/"
             "client-authentication/supported-authentication-methods";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_ch_endpt_ssh_auth_methods_cb);
+#endif
 
+#ifdef NC_ENABLED_TLS
     /* subscribe for TLS Call Home configuration changes */
     xpath = "/ietf-netconf-server:netconf-server/call-home/netconf-client/endpoints/endpoint/tls";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_ch_client_endpt_tls_cb);
@@ -880,7 +918,9 @@ server_data_subscribe(void)
     xpath = "/ietf-netconf-server:netconf-server/call-home/netconf-client/endpoints/endpoint/tls/tls-server-parameters/"
             "client-authentication/cert-maps";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_ch_client_endpt_tls_client_ctn_cb);
+#endif
 
+#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
     /*
      * ietf-keystore (just for in-use operational data)
      */
@@ -894,6 +934,7 @@ server_data_subscribe(void)
     mod_name = "ietf-truststore";
     xpath = "/ietf-truststore:truststore/certificates";
     SR_CONFIG_SUBSCR(mod_name, xpath, np2srv_dummy_cb);
+#endif
 
     /*
      * ietf-netconf-acm
@@ -932,10 +973,12 @@ static void *
 worker_thread(void *arg)
 {
     NC_MSG_TYPE msgtype;
-    int rc, idx = *((int *)arg), monitored;
+    int rc, idx = *((int *)arg);
     struct nc_session *ncs;
 
+#ifdef NC_ENABLED_SSH
     nc_libssh_thread_verbosity(np2_libssh_verbose_level);
+#endif
 
     while (ATOMIC_LOAD_RELAXED(loop_continue)) {
         /* try to accept new NETCONF sessions */
@@ -956,60 +999,41 @@ worker_thread(void *arg)
             continue;
         }
 
-        switch (nc_session_get_ti(ncs)) {
-#ifdef NC_ENABLED_SSH
-        case NC_TI_LIBSSH:
-#endif
-#ifdef NC_ENABLED_TLS
-        case NC_TI_OPENSSL:
-#endif
-#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
-            monitored = 1;
-            break;
-#endif
-        default:
-            monitored = 0;
-            break;
-        }
-
         /* process the result of nc_ps_poll(), increase counters */
         if (rc & NC_PSPOLL_BAD_RPC) {
-            if (monitored) {
-                ncm_session_bad_rpc(ncs);
-            }
+            ncm_session_bad_rpc(ncs);
             VRB("Session %d: thread %d event bad RPC.", nc_session_get_id(ncs), idx);
         }
         if (rc & NC_PSPOLL_RPC) {
-            if (monitored) {
-                ncm_session_rpc(ncs);
-            }
+            ncm_session_rpc(ncs);
             VRB("Session %d: thread %d event new RPC.", nc_session_get_id(ncs), idx);
         }
         if (rc & NC_PSPOLL_REPLY_ERROR) {
-            if (monitored) {
-                ncm_session_rpc_reply_error(ncs);
-            }
+            ncm_session_rpc_reply_error(ncs);
             VRB("Session %d: thread %d event reply error.", nc_session_get_id(ncs), idx);
         }
         if (rc & NC_PSPOLL_SESSION_TERM) {
             VRB("Session %d: thread %d event session terminated.", nc_session_get_id(ncs), idx);
             np2srv_del_session_cb(ncs);
-        } else if (rc & NC_PSPOLL_SSH_CHANNEL) {
+        }
+#ifdef NC_ENABLED_SSH
+        else if (rc & NC_PSPOLL_SSH_CHANNEL) {
             /* a new SSH channel on existing session was created */
             VRB("Session %d: thread %d event new SSH channel.", nc_session_get_id(ncs), idx);
             msgtype = nc_session_accept_ssh_channel(ncs, &ncs);
             if (msgtype == NC_MSG_HELLO) {
                 np2srv_new_session_cb(NULL, ncs);
             } else if (msgtype == NC_MSG_BAD_HELLO) {
-                if (monitored) {
-                    ncm_bad_hello();
-                }
+                ncm_bad_hello(ncs);
             }
         }
+#endif
     }
 
     /* cleanup */
+#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
     nc_thread_destroy();
+#endif
     free(arg);
     return NULL;
 }
@@ -1024,25 +1048,32 @@ print_version(void)
 static void
 print_usage(char* progname)
 {
-    fprintf(stdout, "Usage: %s [-dhV] [-U (path)] [-m mode] [-u uid] [-g gid] [-v level] [-c category]\n", progname);
-    fprintf(stdout, " -d        debug mode (do not daemonize and print verbose messages to stderr instead of syslog)\n");
-    fprintf(stdout, " -h        display help\n");
-    fprintf(stdout, " -V        show program version\n");
-    fprintf(stdout, " -U (path) listen on a local UNIX socket (specific path, default is \"%s\")\n", NP2SRV_UNIX_SOCK_PATH);
-    fprintf(stdout, " -m mode   set mode for the listening UNIX socket\n");
-    fprintf(stdout, " -u uid    set UID/user for the listening UNIX socket\n");
-    fprintf(stdout, " -g gid    set GID/group for the listening UNIX socket\n");
-    fprintf(stdout, " -v level  verbose output level:\n");
-    fprintf(stdout, "               0 - errors\n");
-    fprintf(stdout, "               1 - errors and warnings\n");
-    fprintf(stdout, "               2 - errors, warnings, and verbose messages\n");
+    fprintf(stdout, "Usage: %s [-dhV] [-p path] [-U (path)] [-m mode] [-u uid] [-g gid] [-t timeout] [-v level] [-c category]\n", progname);
+    fprintf(stdout, " -d         debug mode (do not daemonize and print verbose messages to stderr instead of syslog)\n");
+    fprintf(stdout, " -h         display help\n");
+    fprintf(stdout, " -V         show program version\n");
+    fprintf(stdout, " -p path    path to pidfile (default path is \"%s\")\n", NP2SRV_PID_FILE_PATH);
+    fprintf(stdout, " -U[path]   listen on a local UNIX socket (default path is \"%s\")\n", NP2SRV_UNIX_SOCK_PATH);
+    fprintf(stdout, " -m mode    set mode for the listening UNIX socket\n");
+    fprintf(stdout, " -u uid     set UID/user for the listening UNIX socket\n");
+    fprintf(stdout, " -g gid     set GID/group for the listening UNIX socket\n");
+    fprintf(stdout, " -t timeout timeout in seconds of all sysrepo functions (applying edit-config, reading data, ...),\n");
+    fprintf(stdout, "            if 0 (default), the default sysrepo timeouts are used\n");
+    fprintf(stdout, " -v level   verbose output level:\n");
+    fprintf(stdout, "                0 - errors\n");
+    fprintf(stdout, "                1 - errors and warnings\n");
+    fprintf(stdout, "                2 - errors, warnings, and verbose messages\n");
 #ifndef NDEBUG
     fprintf(stdout, " -c category[,category]*\n");
-    fprintf(stdout, "           verbose debug level, print only these debug message categories\n");
-    fprintf(stdout, "           categories: DICT, YANG, YIN, XPATH, DIFF, MSG, LN2DBG, SSH, SYSREPO\n");
+    fprintf(stdout, "            verbose debug level, print only these debug message categories\n");
+# ifdef NC_ENABLED_SSH
+    fprintf(stdout, "            categories: DICT, YANG, YIN, XPATH, DIFF, MSG, LN2DBG, SSH, SYSREPO\n");
+# else
+    fprintf(stdout, "            categories: DICT, YANG, YIN, XPATH, DIFF, MSG, LN2DBG, SYSREPO\n");
+# endif
 #else
     fprintf(stdout, " -c category[,category]*\n");
-    fprintf(stdout, "           verbose debug level, NOT SUPPORTED in release build type\n");
+    fprintf(stdout, "            verbose debug level, NOT SUPPORTED in release build type\n");
 #endif
     fprintf(stdout, "\n");
 }
@@ -1054,6 +1085,7 @@ main(int argc, char *argv[])
     int c, *idx, i;
     int daemonize = 1, verb = 0;
     int pidfd;
+    const char *pidfile = NP2SRV_PID_FILE_PATH;
     char pid[8];
     char *ptr;
     struct passwd *pwd;
@@ -1067,7 +1099,7 @@ main(int argc, char *argv[])
     np2_stderr_log = 1;
 
     /* process command line options */
-    while ((c = getopt(argc, argv, "dhVU::m:u:g:v:c:")) != -1) {
+    while ((c = getopt(argc, argv, "dhVp:U::m:u:g:t:v:c:")) != -1) {
         switch (c) {
         case 'd':
             daemonize = 0;
@@ -1101,11 +1133,16 @@ main(int argc, char *argv[])
             }
 
             nc_verbosity(np2_verbose_level);
+#ifdef NC_ENABLED_SSH
             nc_libssh_thread_verbosity(np2_libssh_verbose_level);
+#endif
             break;
         case 'V':
             print_version();
             return EXIT_SUCCESS;
+        case 'p':
+            pidfile = optarg;
+            break;
         case 'U':
             np2srv.unix_path = optarg ? optarg : NP2SRV_UNIX_SOCK_PATH;
             break;
@@ -1138,6 +1175,16 @@ main(int argc, char *argv[])
                 np2srv.unix_gid = grp->gr_gid;
             }
             break;
+        case 't':
+            np2srv.sr_timeout = strtoul(optarg, &ptr, 10);
+            if (*ptr) {
+                ERR("Invalid timeout value \"%s\".", optarg);
+                return EXIT_FAILURE;
+            }
+
+            /* make ms from s */
+            np2srv.sr_timeout *= 1000;
+            break;
         case 'c':
 #ifndef NDEBUG
             if (verb) {
@@ -1166,9 +1213,11 @@ main(int argc, char *argv[])
                     np2_verbose_level = NC_VERB_DEBUG;
                 } else if (!strcmp(ptr, "LN2DBG")) {
                     np2_verbose_level = NC_VERB_DEBUG_LOWLVL;
+# ifdef NC_ENABLED_SSH
                 } else if (!strcmp(ptr, "SSH")) {
                     /* 2 should be always enough, 3 is too much useless info */
                     np2_libssh_verbose_level = 2;
+# endif
                 } else if (!strcmp(ptr, "SYSREPO")) {
                     np2_sr_verbose_level = SR_LL_DBG;
                 } else {
@@ -1178,7 +1227,9 @@ main(int argc, char *argv[])
             } while ((ptr = strtok(NULL, ",")));
             /* set final verbosity */
             nc_verbosity(np2_verbose_level);
+# ifdef NC_ENABLED_SSH
             nc_libssh_thread_verbosity(np2_libssh_verbose_level);
+# endif
             if (verb) {
                 ly_verb(LY_LLDBG);
                 ly_verb_dbg(verb);
@@ -1209,9 +1260,9 @@ main(int argc, char *argv[])
     }
 
     /* make sure we are the only instance - lock the PID file and write the PID */
-    pidfd = open(NP2SRV_PID_FILE_PATH, O_RDWR | O_CREAT, 0640);
+    pidfd = open(pidfile, O_RDWR | O_CREAT, 0640);
     if (pidfd < 0) {
-        ERR("Unable to open the PID file \"%s\" (%s).", NP2SRV_PID_FILE_PATH, strerror(errno));
+        ERR("Unable to open the PID file \"%s\" (%s).", pidfile, strerror(errno));
         return EXIT_FAILURE;
     }
     if (lockf(pidfd, F_TLOCK, 0) < 0) {
@@ -1219,7 +1270,7 @@ main(int argc, char *argv[])
         if (errno == EACCES || errno == EAGAIN) {
             ERR("Another instance of the Netopeer2 server is running.");
         } else {
-            ERR("Unable to lock the PID file \"%s\" (%s).", NP2SRV_PID_FILE_PATH, strerror(errno));
+            ERR("Unable to lock the PID file \"%s\" (%s).", pidfile, strerror(errno));
         }
         return EXIT_FAILURE;
     }
@@ -1301,8 +1352,10 @@ cleanup:
     sr_unsubscribe(np2srv.sr_data_sub);
     sr_unsubscribe(np2srv.sr_notif_sub);
 
+#if defined(NC_ENABLED_SSH) || defined(NC_ENABLED_TLS)
     /* remove all CH clients so they do not reconnect */
     nc_server_ch_del_client(NULL);
+#endif
 
     /* close all open sessions */
     if (np2srv.nc_ps) {
